/*================================================================================*
                     Inertia Realtime 3D Rendering Engine
   Copyright (c) 1996,97,98, Alex Chalfin, Jeroen Bouwens. All Rights Reserved.
           DISTRIBUTION OF THIS SOURCE CODE IS STRICTLY PROHIBITED
 *================================================================================*/

#ifndef __TYPES_H__
#define __TYPES_H__

//*================================================================================*
// Array size constants
//*================================================================================*
#define MAXTEXTURES 128      // maximum number of textures the engine can handle
#define MAXLIGHTS   8        // 8 light sources 

//*================================================================================*
// Error constants
//*================================================================================*
#define I_OK        0
#define I_MEMERR    1
#define I_FILEERR   2
#define I_FORMATERR 3

//*================================================================================*
// Shading constants
//*================================================================================*
#define AMBIENT       0x00
#define LAMBERT       0x01
#define GOURAUD       0x02
#define PHONG         0x03
#define UNSHADED      0x04

//*================================================================================*
// Surface constants
//*================================================================================*
#define SMOOTH        0x00
#define TEXTURE       0x10
#define REFLECTION    0x20
#define PERSPTEXTURE  0x30

//*================================================================================*
// Transparent constant
//*================================================================================*
#define TRANSPARENT   0x80


//*================================================================================*
// RGB Type for palette. Packed to byte boundary
//*================================================================================*
#pragma pack (1)
typedef struct
{
	char r, g, b;
} RGBType;

//*================================================================================*
// Palette and Matrix types
//*================================================================================*
typedef RGBType Palette[256];
typedef float Matrix4x3[12];

//*================================================================================*
// Texture declaration
//*================================================================================*
typedef struct
{
	char    *Texture;
	Palette Pal;
} TextureRec;

//*================================================================================*
// 3D Vector/Vertex declaration
//*================================================================================*
typedef struct
{
	float x, y, z;
} Vertex3d;

//*================================================================================*
// 2D Vector/Vertex declaration
//*================================================================================*
typedef struct
{
	float x, y;
} Vertex2d;

//*================================================================================*
// Vertex declartion (local, camera, screen space, etc...)
//*================================================================================*
typedef struct
{
	Vertex3d Loc;		// Local space vector
	Vertex3d Cam;		// Camera space vector
	Vertex2d Scr;		// Screen space vector
	Vertex2d Env;		// Environment mapping coordiante
	Vertex3d N;			// Vertex normal
	float    OneOverZ;	// 1/Z value (calculated in 3D->2D projection)
	float    Shade;		// shade/intensity of vertex
	int      Visible;	// visibility flag
} Vertex;

//*================================================================================*
// Triangle declartion (index face set triangle)
//*================================================================================*
typedef struct
{
	Vertex   *A;		// Vertex A
	Vertex   *B;		// Vertex B
	Vertex   *C;		// Vertex C
	Vertex3d  N;		// Polygon normal
	Vertex2d  UV1;		// Texture coord for vertex A
	Vertex2d  UV2;		// Texture coord for vertex B
	Vertex2d  UV3;		// Texture coord for vertex C
	char     *Texture;	// Texture to use for this polygon
	char     *IBuffer;	// Intensity buffer to use for this polygon
	float     Shade;	// Shade of the polygon
	int       Visible;	// polygon visibility flag
	int       Color;	// color of the polygon
} Triangle;


//*================================================================================*
// Renderlist node structure (for sorting)
//*================================================================================*
typedef struct _tSortRec
{
	int              ZMid;			// Midpoint Z of the triangle
	int              ShadeSurface;	// Rendermask for polygon
	struct _tSortRec *Next;			// Linked list next pointer
	Triangle         *Poly;			// pointer to the actual polygon
} SortRec;

//*================================================================================*
// Renderlist declartion
//*================================================================================*
typedef struct
{
	int       Allocated;	// Number of allocated sortlist emelments
	SortRec  *SortList;		// list of elements
	int       Count;		// number of elements currently in the list
	SortRec  *MSBList[256];	// radix sorting list A
	SortRec  *LSBList[256];	// redix sorting list B
} tRenderList;

//*================================================================================*
// Light list declartion 
//*================================================================================*
typedef struct
{
	int NumLights;					// number of lights used
	Vertex3d Light[MAXLIGHTS];		// list of lights in world space
	Vertex3d TempLight[MAXLIGHTS];	// list of lights in temporary space
} tLightList;

//*================================================================================*
// VectorObject declartion
//*================================================================================*
typedef struct
{
	int       NumVert;		// Number of vertices in the object
	int       NumPoly;		// Number of polygons in the object
	Vertex   *Vertexs;		// list of vertices
	Triangle *Polygons;		// list of triangles

	int       RenderMask;	// current rendering mask of the object
	float     Radius;		// radius of the object's boundary sphere
	int       CorrectCount;	// orthonormalization count for matrix
	int       ClipMask;		// object clipping mask (dynamic)
	int       Cull;			// object culling flag
	Matrix4x3 Orientation;	// orientation matrix 
} VectorObject;

//*================================================================================*
// ViewObject declartion
//*================================================================================*
typedef struct
{
	float CenterX;			// X center of projection
	float CenterY;			// Y center of projection

	int	  WinMinX, WinMinY;	// Viewport dimentions
	int   WinMaxX, WinMaxY;	
	
	float InvCamZFac;		// Inverse field of view parameter
	float InvCamYFac;		// Inverse x/y aspect ratio parameter
	float CamZFac;			// Field of view parameter
	float CamYFac;			// x/y aspect ratio parameter

	float FocalLength;		// current focal length (mm)
	float FieldOfView;		// field of view (in degrees)

	float FustrumMinZ;		// near clipping plane
	float FustrumMaxZ;		// far clipping plane

	Matrix4x3  Camera;		// orientation matrix for camera
} ViewObject;


//*================================================================================*
// Stuff that doesn't need to be used but is none the less global
//*================================================================================*

#include "define.inc"

#ifdef  SHAREWARE
#define MAX_POLYGONS   500   // shareware polygon limit
#endif

#define CORRECTRES  32       // orthonormalize rotation matrices every 32 frames

#define XLOC        9        // index in matrix for X location 
#define YLOC        10       // index in matrix for Y location 
#define ZLOC        11       // index in matrix for Z location 

//*================================================================================*
// Shade and Surface masks
//*================================================================================*
#define SHADE_MASK    0x0f
#define SURFACE_MASK  0x70

//*================================================================================*
// Clipping flags (used inertnally)
//*================================================================================*
#define CLIP_Z        0x01
#define CLIP_LEFT     0x02
#define CLIP_RIGHT    0x04
#define CLIP_TOP      0x08
#define CLIP_BOTTOM   0x10
#define CLIP_GOURAUD  0x20
#define CLIP_TEXTURE  0x40
#define CLIP_ENV      0x80

#define CLIP_SHADEMASK  (CLIP_GOURAUD | CLIP_TEXTURE | CLIP_ENV)

typedef void (*TriangleFiller)(Triangle *P, char *VPage, int x, int y);
typedef void (*ShaderFunction)(ViewObject *View, VectorObject *Obj, float *Matrix);


#endif
